class Node:
    def __init__(self, prev):
        self.next = None
        if prev:
            prev.next = self
        self.jumpspot = False
        self.prev = prev
        self.line = 0

    def set_line(self, n):
        self.line = n

    def execute(self, executor):
        executor.set_cur(self.next)


class Container:
    def __init__(self, state):
        self.state = state
        self.name = None

    def __eq__(self, other):
        return self.state == other.state
    
    def __repr__(self):
        if self.name:
            return self.name
        return "".join(("cont", repr(self.state)))
    
    def set_name(self, name):
        self.name = name

    def change_to(self, cont):
        if isinstance(cont, Container):
            self.state = cont.state
        else:
            self.state = cont
        self.graph.update()


class IfNode(Node):
    args = ((Container, "equal"), (Container, "equal", "different"))
    color = (0, 0, 255)
    def __init__(self, prev, a, b):
        super().__init__(prev)
        self.a = a
        self.b = b
        self.inext = None

    def __repr__(self):
        #if not self.inext.line:
        #    print(":::  ", self.inext.prev, self.inext.mark)
        #return " ".join((repr(self.line), repr(self.mark), "if", repr(self.a), "==",
        #                 repr(self.b), ":", repr(self.inext.line)))
        return "({}) if {} == {}: {}".format(self.line, self.a, self.b, self.inext.line)
    
    def visual(self):
        return {"arrows to": [self.inext],
                "c. arrows from": [self.a, self.b]}

    def set_inext(self, inext):
        self.inext = inext
        inext.jumpspot = True

    def execute(self, executor):
        if self.a == self.b:
            executor.set_cur(self.inext)
        else:
            super().execute(executor)


class Mover(Node):
    args = ((Container, "equal"), (Container, "different"))
    color = (0, 255, 0)
    def __init__(self, prev, source, dest):
        super().__init__(prev)
        self.source = source
        self.dest = dest

    def __repr__(self):
        return "({}) move {} to {}".format(self.line, self.source, self.dest)

    def visual(self):
        return {"c. arrows to": [self.dest],
                "c. arrows from": [self.source]}

    def execute(self, executor):
        self.dest.change_to(self.source)
        super().execute(executor)


class Waiter(Node):
    args = ((int,),)
    color = (255, 0, 0)
    def __init__(self, prev, n=1):
        super().__init__(prev)
        self.n = n
        self.countdown = n

    def __repr__(self):
        return "({}) wait {}".format(self.line, self.n)
    
    def visual(self):
        return {"text": str(self.n)}

    def execute(self, executor):
        if self.countdown:
            self.countdown -= 1
            executor.wait()
        else:
            self.countdown = self.n
            super().execute(executor)

def sanitize(x):
    return min(1, max(-1, x))

class Add(Node):
    args = ((Container, "equal"), (Container, "different"))
    color = (200, 0, 200)
    def __init__(self, prev, source, dest):
        super().__init__(prev)
        self.source = source
        self.dest = dest

    def __repr__(self):
        return "({}) add {} to {}".format(self.line, self.source, self.dest)

    def visual(self):
        return {"c. arrows to": [self.dest],
                "c. arrows from": [self.source]}

    def execute(self, executor):
        x1, y1 = self.source.state
        x2, y2 = self.dest.state
        self.dest.change_to((sanitize(x1 + x2), sanitize(y1 + y2)))
        super().execute(executor)


class Rotator(Node):
    args = ((str,), (Container, "equal"))
    color = (0, 255, 255)
    seq = [(0, -1), (1, -1), (1, 0), (1, 1), (0, 1), (-1, 1), (-1, 0),
            (-1, -1)]
    def __init__(self, prev, dir_, cont):
        super().__init__(prev)
        self.dir = dir_
        self.cont = cont

    def __repr__(self):
        return "({}) rotate {} to {}".format(self.line, self.cont, self.dir)

    def visual(self):
        return {"text": self.dir,
                "c. arrows to": [self.cont]}

    def execute(self, executor):
        x, y = self.cont.state
        if x or y:
            i = self.seq.index((x, y))
            if self.dir == "R":
                i += 1
            else:
                i -= 1
            self.cont.change_to(self.seq[i % 8])
        super().execute(executor)


class Executor:
    def __init__(self, start):
        self.start = start
        self.cur = start
        self.must_wait = False

    def set_cur(self, node):
        self.cur = node

    def wait(self):
        self.must_wait = True

    def execute(self):
        while self.cur is not None:
            self.cur.execute(self)
            if self.must_wait:
                self.must_wait = False
                yield self.cur
        self.cur = self.start
        self.must_wait = False
        yield None


def execute(*exes):
    generators = [exe.execute() for exe in exes]
    while True:
        for n, exe in enumerate(generators):
            ret = next(exe)
            if ret is None:
                generators[n] = exes[n].execute()
                ret = next(generators[n])
            yield ret


if __name__ == "__main__":
    state = Container((0, 0, 0))
    const_1 = Container((0, 1, 0))
    const_2 = Container((1, 0, 0))
    nodes = []
    nodes.append(Mover(None, const_1, state))
    nodes.append(Waiter(nodes[-1], 5))
    nodes.append(Mover(nodes[-1], const_2, state))
    nodes.append(Waiter(nodes[-1]))
    exe = Executor(nodes[0])
    for node in exe.execute():
        print(state.state)
